from unittest import TestCase

import numpy as np
from numpy.testing import assert_almost_equal
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.restapi.app import app
from omegaml.restapi.tests.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class OmegaRestAsyncApiTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.client = RequestsLikeTestClient(app)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    @property
    def _headers(self):
        return {}

    def test_predict_from_data_inline(self):
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        self.om.models.put(clf, 'regression')
        # check we can use it to predict
        resp = self.client.put('/api/v1/model/regression/predict', json={
            'columns': ['v'],
            'data': dict(v=[5]),
        }, auth=self.auth, headers=self._async_headers)
        resp = self._check_async(resp)
        data = resp.get_json()['response']  # prediction result
        self.assertEqual(data.get('model'), 'regression')
        self.assertEqual(data.get('result'), [10.])

    def test_predict_from_dataset(self):
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        self.om.models.put(clf, 'regression')
        self.om.datasets.put([5], 'foo', append=False)
        # check we can use it to predict
        resp = self.client.put('/api/v1/model/regression/predict?datax=foo',
                               json={}, auth=self.auth, headers=self._async_headers)
        resp = self._check_async(resp)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()['response']
        self.assertEqual(data.get('model'), 'regression')
        self.assertEqual(data.get('result'), [10.])

    def test_predict_from_data_inline_versions(self):
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        self.om.models.put(clf, 'regression', tag='commit1')
        clf.intercept_ = 10
        self.om.models.put(clf, 'regression', tag='commit2')
        # check we can use it to predict previous version
        resp = self.client.put('/api/v1/model/regression^/predict', json={
            'columns': ['v'],
            'data': dict(v=[5]),
        }, auth=self.auth, headers=self._async_headers)
        resp = self._check_async(resp)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()['response']
        self.assertEqual(data.get('model'), 'regression^')
        assert_almost_equal(data.get('result'), [10.])
        # check we can use it to predict current version
        resp = self.client.put('/api/v1/model/regression/predict', json={
            'columns': ['v'],
            'data': dict(v=[5]),
        }, auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertEqual(data.get('model'), 'regression')
        assert_almost_equal(data.get('result'), [20.])
        # check we can use it to predict tagged version
        resp = self.client.put('/api/v1/model/regression@commit1/predict', json={
            'columns': ['v'],
            'data': dict(v=[5]),
        }, auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertEqual(data.get('model'), 'regression@commit1')
        assert_almost_equal(data.get('result'), [10.])




import os

import unittest
from pandas.util.testing import assert_frame_equal

from omegaml import Omega
import pandas as pd

from omegaml.tests.util import OmegaTestMixin


class IOToolsMixinTests(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()

    def test_writecsv(self):
        om = self.om
        df = pd.DataFrame({
            'x': range(10)
        })
        om.datasets.put(df, 'testdf')
        fn = '/tmp/testdf.csv'
        os.remove(fn) if os.path.exists(fn) else None
        om.datasets.to_csv('testdf', fn, index=False)
        self.assertTrue(os.path.exists(fn))
        dfx = pd.read_csv(fn)
        assert_frame_equal(dfx, df)

    def test_writecsv_with_apply(self):
        om = self.om
        df = pd.DataFrame({
            'x': range(10)
        })
        def myfunc(df):
            df['y'] = df['x'] * 2
        om.datasets.put(df, 'testdf')
        fn = '/tmp/testdf.csv'
        os.remove(fn) if os.path.exists(fn) else None
        om.datasets.to_csv('testdf', fn, apply=myfunc)
        self.assertTrue(os.path.exists(fn))
        dfx = pd.read_csv(fn)
        self.assertIn('y', dfx.columns)

    def test_readcsv(self):
        om = self.om
        df = pd.DataFrame({
            'x': range(10)
        })
        om.datasets.put(df, 'testdf')
        fn = '/tmp/testdf.csv'
        os.remove(fn) if os.path.exists(fn) else None
        om.datasets.to_csv('testdf', fn)
        self.assertTrue(os.path.exists(fn))






import mimetypes
from base64 import encodebytes, decodebytes

import json
import nbformat
import os
from datetime import datetime
from io import BytesIO
from notebook.base.handlers import HTTPError
from notebook.services.contents.manager import ContentsManager
from tornado import web
from urllib.parse import unquote

from omegaml.notebook.checkpoints import NoOpCheckpoints


class OmegaStoreContentsManager(ContentsManager):
    """
    Jupyter notebook storage manager for omegaml

    Adopted from notebook/services/contents/filemanager.py

    This requires a properly configured omegaml instance.
    see http://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html
    """

    def __init__(self, **kwargs):
        # pass omega= for testing purpose
        self._omega = kwargs.pop('omega', None)
        super(OmegaStoreContentsManager, self).__init__(**kwargs)

    def _checkpoints_class_default(self):
        return NoOpCheckpoints

    @property
    def omega(self):
        """
        return the omega instance used by the contents manager
        """
        if self._omega is None:
            import omegaml as om
            self._omega = om
        self._omega.jobs._include_dir_placeholder = True
        return self._omega

    @property
    def store(self):
        """
        return the OmageStore for jobs (notebooks)
        """
        return self.omega.jobs.store

    @property
    def _dir_placeholder(self):
        return self.omega.jobs._dir_placeholder

    def get(self, path, content=True, type=None, format=None):
        """
        get an entry in the store

        this is called by the contents engine to get the contents of the jobs
        store.
        """
        path = unquote(path).strip('/')
        if type == 'notebook' or (type is None and path.endswith('.ipynb')):
            model = self._notebook_model(path, content=content)
        elif type == 'file':
            model = self._file_model(path, content=content)
        elif type in (None, 'directory'):
            # jupyterlab passes None to get directory
            # we never return content with a directory listing to save time
            # the frontend will request the specific contents
            model = self._dir_model(path, content=False)
        else:
            raise web.HTTPError(404, u'Type {} at {} is not supported'.format(type, path))
        return model

    def save(self, model, path):
        """
        save an entry in the store

        this is called by the contents engine to store a notebook
        """
        om = self.omega
        path = unquote(path).strip('/')
        type = model.get('type')
        name = model.get('name')
        self.run_pre_save_hook(model=model, path=path)
        if type is None:
            raise web.HTTPError(400, u'No file type provided')
        try:
            if type == 'notebook' or (type == 'file' and path.endswith('.ipynb')):
                content = model.get('content')
                # parse input
                if model.get('format') == 'text' and isinstance(content, bytes):
                    content = content.decode('utf8')
                elif model.get('format') == 'base64':
                    if isinstance(content, str):
                        content = content.encode('ascii')
                    content = decodebytes(content).decode('utf8')
                while isinstance(content, str):
                    content = json.loads(content)
                    model['content'] = content
                    model['format'] = None
                if content is None or not isinstance(content, dict):
                    raise web.HTTPError(400, u'No file content provided or wrong format')
                # create notebook
                nb = nbformat.from_dict(content)
                self.check_and_sign(nb, path)
                self.omega.jobs.put(nb, path)
                self.validate_notebook_model(model)
                validation_message = model.get('message', None)
                model = self.get(path, content=False, type=type)
                if validation_message:
                    model['message'] = validation_message
            elif type == 'directory':
                ph_name = '{path}/{self._dir_placeholder}'.format(**locals()).strip('/')
                self.omega.jobs.create("#placeholder", ph_name)
                model = self.get(path, content=False, type=type)
                model['content'] = None
                model['format'] = None
                validation_message = None
            elif type == 'file':
                content = model.get('content')
                fmt = model.get('format')
                self._save_file(path, content, fmt)
                model = self.get(path, content=False, type=type)
            else:
                raise web.HTTPError(
                    400, "Unhandled contents type: %s" % model['type'])
        except web.HTTPError:
            raise
        except Exception as e:
            self.log.error(
                u'Error while saving file: %s %s', path, e, exc_info=True)
            raise web.HTTPError(
                500, u'Unexpected error while saving file: %s %s' % (path, e))
        return model

    def delete_file(self, path):
        """
        delete an entry

        this is called by the contents engine to delete an entry
        """
        path = unquote(path).strip('/')
        try:
            self.omega.jobs.drop(path)
        except Exception as e:
            self.omega.jobs.drop(path + '/' + self._dir_placeholder)

    def rename_file(self, old_path, new_path):
        """
        rename a file

        this is called by the contents engine to rename an entry
        """
        old_path = unquote(old_path).strip('/')
        new_path = unquote(new_path).strip('/')
        # check file or directory
        if self.file_exists(new_path):
            raise web.HTTPError(409, u'Notebook already exists: %s' % new_path)
        elif self.dir_exists(new_path):
            raise web.HTTPError(409, u'Directory already exists: %s' % new_path)
        # do the renaming
        if self.dir_exists(old_path):
            old_dirname = old_path + '/' + self._dir_placeholder
            new_dirname = new_path + '/' + self._dir_placeholder
            meta = self.omega.jobs.metadata(old_dirname)
            meta.name = new_dirname
        elif self.file_exists(old_path):
            meta = self.omega.jobs.metadata(old_path)
            meta.name = new_path
        # rename on metadata. Note the gridfile instance stays the same
        meta.save()

    def exists(self, path):
        """
        Does a file or dir exist at the given collection in gridFS?
        We do not have dir so dir_exists returns true.

        :param path: (str) The relative path to the file's directory
          (with '/' as separator)
        :returns exists: (boo) The relative path to the file's directory (with '/' as separator)
        """
        path = unquote(path).strip('/')
        return self.file_exists(path) or self.dir_exists(path)

    def dir_exists(self, path=''):
        """check if directory exists

        Args:
            path: name of directory

        Returns:
            True if directory exists
        """
        path = unquote(path).strip('/')
        if path == '':
            return True
        pattern = r'^{path}.*/({placeholder}|.+)'.format(path=path, placeholder=self._dir_placeholder)
        return len(self.omega.jobs.list(regexp=pattern)) > 0

    def file_exists(self, path):
        """check if file exists

        Args:
            path: name of file

        Returns:
            True if file exists
        """
        path = unquote(path).strip('/')
        if not path:
            return False
        # always check for an actual file, not some sub path
        pattern = r'^{}$'.format(path)
        does_exist = len(self.omega.jobs.list(regexp=pattern)) > 0
        does_exist |= len(self.omega.datasets.list(regexp=pattern)) > 0
        return does_exist

    def is_hidden(self, path):
        """check if path or file is hidden

        Args:
            path: name of file or path

        Returns:
            False, currently always returns false
        """
        return False

    def _read_notebook(self, path, as_version=None):
        path = unquote(path).strip('/')
        return self.omega.jobs.get(path)

    def _notebook_model(self, path, content=True, meta=None):
        """
        Build a notebook model
        if content is requested, the notebook content will be populated
        as a JSON structure (not double-serialized)
        """
        path = unquote(path).strip('/')
        model = self._base_model(path)
        model['type'] = 'notebook'
        # always add accurate created and modified
        meta = meta or self.omega.jobs.metadata(path)
        if meta is not None:
            model['created'] = meta.created
            model['last_modified'] = meta.modified
        if content:
            nb = self._read_notebook(path, as_version=4)
            if nb is None:
                raise HTTPError(400, "Cannot read non-file {}".format(path))
            self.mark_trusted_cells(nb, path)
            model['content'] = nb
            model['format'] = 'json'
            self.validate_notebook_model(model)
        return model

    def _base_model(self, path, kind=None):
        """Build the common base of a contents model"""
        # http://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html
        path = unquote(path).strip('/')
        last_modified = datetime.utcnow()
        created = last_modified
        # Create the base model.
        model = {}
        model['name'] = os.path.basename(path)
        model['path'] = path
        model['last_modified'] = last_modified
        model['created'] = created
        model['content'] = None
        model['format'] = None
        model['mimetype'] = None
        model['writable'] = True
        if kind:
            model['type'] = kind
            model['content'] = [] if kind == 'directory' else None
        return model

    def _dir_model(self, path, content=True):
        """
        Build a model to return all of the files in gridfs
        if content is requested, will include a listing of the directory
        """
        # this looks like a seemingly simple task, it's carefully crafted
        path = unquote(path).strip('/')
        model = self._base_model(path, kind='directory')
        model['format'] = 'json'
        contents = model['content']
        # get existing entries from a pattern that matches either
        #    top-level files: ([\w -]+\.[\w -]*)
        #    directories (files in): ([\w ]+/([\w ]+\.[\w]*))
        # does not work:
        #    pattern = r'([\w ]+/_placeholder\.[\w]*)|([\w ]+\.[\w]*)$'
        #    it is too restrictive as entries can be generated without a placeholder
        # so we get all, which may include sub/sub/file
        # and we need to include them because we need to find sub/sub directories
        # note \w is any word character (letter, digit, underscore)
        #      \s is any white space
        #      \d is any digit
        #      :_  match literally
        #      [^\/]  matches any character except /
        #pattern = r'([\w\s\-.\d:()+]+\/)?([\w\s\-.\d:()+]+\.[\w]*)$'
        pattern = r'([^\/]+\/)?([^\/]+\.[^\/]*)$'
        # if we're looking in an existing directory, prepend that
        if path:
            pattern = r'{path}/{pattern}'.format(path=path, pattern=pattern)
        pattern = r'^{}'.format(pattern)
        entries = self.omega.jobs.list(regexp=pattern, raw=True)
        if path and not entries:
            raise HTTPError(400, "Directory not found {}".format(path))
        # by default assume the current path is listed already
        directories = [path]
        for meta in entries:
            # get path of entry, e.g. sub/foo.ipynb => sub
            entry_path = os.path.dirname(meta.name)
            # if not part of listed directories yet, include
            if entry_path not in directories:
                entry = self._base_model(entry_path, kind='directory')
                contents.append(entry)
                directories.append(entry_path)
            # ignore placeholder files
            if meta.name.endswith(self._dir_placeholder):
                continue
            # only include files that are in the path we're listing
            if entry_path != path:
                continue
            # include the actual file
            try:
                entry = self._notebook_model(meta.name, content=content, meta=meta)
            except Exception as e:
                msg = ('_dir_model error, cannot get {}, '
                       'removing from list, exception {}'.format(meta.name, str(e)))
                self.log.warning(msg)
            else:
                contents.append(entry)
        return model

    def _file_model(self, path, content=True, format=None):
        model = self._base_model(path)
        model['type'] = 'file'

        model['mimetype'] = mimetypes.guess_type(path)[0]

        if content:
            content, format = self._read_file(path, format)
            if model['mimetype'] is None:
                default_mime = {
                    'text': 'text/plain',
                    'base64': 'application/octet-stream'
                }[format]
                model['mimetype'] = default_mime

            model.update(
                content=content,
                format=format,
            )

        return model

    def _read_file(self, os_path, format):
        """Read a non-notebook file.

        os_path: The path to be read.
        format:
          If 'text', the contents will be decoded as UTF-8.
          If 'base64', the raw bytes contents will be encoded as base64.
          If not specified, try to decode as UTF-8, and fall back to base64
        """
        if os_path.endswith('.ipynb'):
            meta = self.omega.jobs.metadata(os_path)
        else:
            meta = self.omega.datasets.metadata(os_path)
        if meta is None or meta.gridfile is None:
            raise HTTPError(400, "Cannot read non-file %s" % os_path)

        if meta.gridfile:
            bcontent = meta.gridfile.read()
            meta.gridfile.close()

        if format is None or format == 'text':
            # Try to interpret as unicode if format is unknown or if unicode
            # was explicitly requested.
            try:
                return bcontent.decode('utf8'), 'text'
            except UnicodeError:
                if format == 'text':
                    raise HTTPError(
                        400,
                        "%s is not UTF-8 encoded" % os_path,
                        reason='bad format',
                    )
        return encodebytes(bcontent).decode('ascii'), 'base64'

    def _save_file(self, os_path, content, format):
        """Save content of a generic file."""
        if format not in {'text', 'base64'}:
            raise HTTPError(
                400,
                "Must specify format of file contents as 'text' or 'base64'",
            )
        try:
            if format == 'text':
                bcontent = content.encode('utf8')
            else:
                b64_bytes = content.encode('ascii')
                bcontent = decodebytes(b64_bytes)
        except Exception as e:
            raise HTTPError(
                400, u'Encoding error saving %s: %s' % (os_path, e)
            )

        self.omega.datasets.put(BytesIO(bcontent), os_path)



import pandas as pd
import requests
from datetime import datetime, timedelta
from tabulate import tabulate
from time import sleep

from omegaml.client.auth import OmegaRestApiAuth
from omegaml.client.docoptparser import CommandBase
from omegaml.client.userconf import save_userconfig_from_apikey
from omegaml.client.util import get_omega


class CloudCommandBase(CommandBase):
    """
    Usage:
      om cloud login [<userid>] [<apikey>] [options]
      om cloud config [show] [options]
      om cloud (add|update|remove) <kind> [--specs <specs>] [options]
      om cloud status [runtime|pods|nodes|dbsize] [options]
      om cloud log <pod> [--since <time>] [options]
      om cloud metrics [<metric_name>] [--since <time>] [--start <start>] [--end <end>] [--step <step>] [--plot] [options]

    Options:
      --userid=USERID   the userid at hub.omegaml.io (see account profile)
      --apikey=APIKEY   the apikey at hub.omegaml.io (see account profile)
      --apiurl=URL      the cloud URL [default: https://hub.omegaml.io]
      --count=NUMBER    how many instances to set up [default: 1]
      --node-type=TYPE  the type of node [default: small]
      --specs=SPECS     the service specifications as "key=value[,...]"
      --since=TIME      recent log time, defaults to 5m (5 minutes)
      --start=DATETIME  start datetime of range query
      --end=DATETIME    end datetime of range query
      --step=UNIT       step in seconds or duration unit (s=seconds, m=minutes)
      --plot            if specified use plotext library to plot (preliminary)

    Description:
      om cloud is available for the omega|ml managed service at https://hub.omegaml.io

      Logging in
      ----------

      $ om cloud login <userid> <apikey>

      Showing the configuration
      -------------------------

      $ om cloud config

      Building a cluster
      ------------------

      Set up a cluster

      $ om cloud add nodepool --specs "node-type=<node-type>,role=worker,size=1"
      $ om cloud add runtime --specs "role=worker,label=worker,size=1"

      Switch nodes on and off

      $ om cloud update worker --specs "node-name=<name>,scale=0" # off
      $ om cloud update worker --specs "node-name=<name>,scale=1" # on

      Using Metrics
      -------------

      The following metrics are available

      * node-cpu-usage      node cpu usage in percent
      * node-memory-usage   node memory usage in percent
      * node-disk-uage      node disk usage in percent
      * pod-cpu-usage       pod cpu usage in percent
      * pod-memory-usage    pod memory usage in bytes

      Get the specific metrics as follows, e.g.

      $ om cloud metrics node-cpu-usage
      $ om cloud metrics pod-cpu-usage --since 30m
      $ om cloud metrics pod-memory-usage --start 20dec2020T0100 --end20dec2020T0800
    """
    command = 'cloud'

    @property
    def om(self):
        without_config =('config', 'login')
        require_config = not any(self.args.get(k) for k in without_config)
        return get_omega(self.args, require_config=require_config)

    def login(self):
        userid = self.args.get('<userid>') or self.args.get('--userid')
        apikey = self.args.get('<apikey>') or self.args.get('--apikey')
        api_url = self.args.get('--apiurl')
        configfile = self.args.get('--config') or 'config.yml'
        if not userid:
            userid = self.ask('Userid:')
        if not apikey:
            apikey = self.ask('Apikey:')
        save_userconfig_from_apikey(configfile, userid, apikey, api_url=api_url)

    def config(self):
        om = self.om
        config_file = om.defaults.OMEGA_CONFIG_FILE
        if config_file is None:
            config_file = print("No configuration file identified, assuming defaults")
        # print config
        restapi_url = getattr(om.defaults, 'OMEGA_RESTAPI_URL', 'not configured')
        runtime_url = om.runtime.celeryapp.conf['BROKER_URL']
        userid = getattr(om.defaults, 'OMEGA_USERID', 'not configured')
        self.logger.info('Config file: {config_file}'.format(**locals()))
        self.logger.info('User id: {userid}'.format(**locals()))
        self.logger.info('REST API URL: {restapi_url}'.format(**locals()))
        self.logger.info('Runtime broker: {runtime_url}'.format(**locals()))

    def add(self):
        command_url = self._issue_command('install')

    def update(self):
        self._issue_command('update')

    def remove(self):
        self._issue_command('uninstall')

    def _issue_command(self, phase):
        om = self.om
        offering = self.args.get('<kind>')
        size = self.args.get('--count')
        node_type = self.args.get('--node-type')
        specs = self.args.get('--specs')
        user = getattr(om.defaults, 'OMEGA_USERID')
        default_specs = "size={size},node-type={node_type}".format(**locals())
        params = specs or default_specs
        data = {
            'offering': offering,
            'user': user,
            'phase': phase,
            'params': params,
        }
        command = self._request_service_api(om, 'post', service='command', data=data)
        self._wait_service_command(om, command)

    def _request_service_api(self, om, method, service=None, data=None, uri=None):
        """
        call a service API for a given omega instance

        Args:
            om (Omega): the omega instance, must be authenticated to the cloud
            method (str): the method verb (CREATE, GET, POST, UPDATE, DELETE)
            service (str): the name of the service endpoint as in /api/service/<name>
            data (dict): the data to send

        Returns:
            The response json
        """
        auth = OmegaRestApiAuth.make_from(om)
        restapi_url = getattr(om.defaults, 'OMEGA_RESTAPI_URL', 'not configured')
        uri = uri or '/admin/api/v2/service/{service}'.format(**locals())
        service_url = '{restapi_url}{uri}'.format(**locals())
        method = getattr(requests, method)
        resp = method(service_url, json=data, auth=auth)
        resp.raise_for_status()
        if resp.status_code == requests.codes.created:
            # always request the actual object
            url = resp.headers['Location']
            resp = requests.get(url, auth=auth)
            resp.raise_for_status()
        return resp.json()

    def _wait_service_command(self, om, command):
        """
        wait until the service command status is completed or failed

        Args:
            command (dict): the command object as returned by the service api

        Returns:

        """
        import tqdm
        with tqdm.tqdm(unit='s', total=30) as progress:
            while True:
                progress.update(1)
                uri = command.get('resource_uri')
                data = self._request_service_api(om, 'get', uri=uri)
                status = data.get('status')
                if int(status) > 1:
                    break
                sleep(1)
        if status == '5':
            self.logger.info("Ok, done.")
        else:
            msg = "Error {status} occurred on {uri}".format(**locals())
            self.logger.error(msg)

    def status(self):
        # self.status_nodes()
        kinds = ('runtime', 'pods', 'nodes', 'dbsize')
        om = self.om
        auth = OmegaRestApiAuth.make_from(om)
        for kind in (filter(lambda k: self.args.get(k), kinds)):
            status_meth = getattr(self, f'status_{kind}')
            status_meth(kind, auth)
            break
        else:
            print(f"status is available for {kinds}")

    def metrics(self):
        # available metrics
        metrics = ('node_cpu_usage', 'node_memory_usage', 'node_disk_usage',
                   'pod_memory_usage', 'pod_cpu_usage')
        # column in prom2df dataframe for given metric group
        metric_group_column = {
            'node': 'node',
            'pod': 'pod_name',
        }
        metric_name = (self.args.get('<metric_name>') or '').replace('-', '_')
        since = self.args.get('--since') or None
        start = self.args.get('--start') or None
        end = self.args.get('--end') or None
        step = self.args.get('--step') or None
        should_plot = self.args.get('--plot')
        # check if we have a valid metric
        if metric_name in metrics:
            # get default range, if any
            if since:
                if any(since.endswith(v) for v in 'hms'):
                    # a relative time spec
                    unit = dict(h='hours', m='minutes', s='seconds').get(since[-1])
                    delta = int(since[0:-1])
                    start = datetime.utcnow() - timedelta(**{unit: delta})
                else:
                    # absolute time
                    start = pd.to_datetime(since, utc=True)
            if start:
                start = pd.to_datetime(start, utc=True)
            if end:
                end = pd.to_datetime(end, utc=True)
            if start and not end:
                end = datetime.utcnow()
            if end and not start:
                start = (end - timedelta(minutes=10)).isoformat()
            if (start or end) and not step:
                step = '5m'
            # query
            om = self.om
            auth = OmegaRestApiAuth.make_from(om)
            data = self._get_metric(metric_name, auth, start=start, end=end, step=step)
            try:
                df = prom2df(data['objects'], metric_name)
            except:
                print("No data could be found. Check the time range.")
                return
            if should_plot:
                import plotext as plx
                # as returned by plx.get_colors
                colors = 'red', 'green', 'yellow', 'organge', 'blue', 'violet', 'cyan'
                metric_group = metric_group_column[metric_name.split('_', 1)[0]]
                for i, (g, gdf) in enumerate(df.groupby(metric_group)):
                    x = range(0, len(gdf))
                    y = gdf['value'].values
                    plx.plot(x, y, line_color=colors[i])
                plx.show()
            else:
                print(tabulate(df, headers='keys'))
        else:
            print("Available metrics:", metrics)

    def _get_metric(self, name, auth, **query):
        url = f'https://hub.omegaml.io/apps/omops/dashboard/api/v1/metrics/{name}'
        resp = requests.get(url, auth=auth, params=query)
        data = resp.json()
        return data

    def _get_status(self, kind, auth):
        url = f'https://hub.omegaml.io/apps/omops/dashboard/api/v1/status/{kind}'
        resp = requests.get(url, auth=auth)
        data = resp.json()
        return data

    def _get_logs(self, podname, since, auth):
        url = f'https://hub.omegaml.io/apps/omops/dashboard/api/v1/logs/{podname}?since={since}'
        resp = requests.get(url, auth=auth)
        data = resp.json()
        return data

    def status_runtime(self, kind, auth):
        data = self._get_status(kind, auth)
        active_tasks = data['objects'][0].get('active') or {}
        queues = data['objects'][0]['queues']
        workers = [{'worker': k,
                    'tasks': len(v),
                    'labels': ','.join(q['name']
                                       for q in queues.get(k, [])
                                       # filter amq internal queues
                                       if not q['name'].startswith('amq.')),
                    } for k, v in active_tasks.items()]
        if workers:
            print(tabulate(workers, headers='keys', showindex=False))
        else:
            print("No runtime workers found.")

    def status_pods(self, kind, auth):
        data = self._get_status(kind, auth)
        pods = data.get('objects')
        df = pd.DataFrame.from_dict(pods)
        print(tabulate(df, headers='keys', showindex=False))

    def status_nodes(self, kind, auth):
        data = self._get_status(kind, auth)
        nodes = data.get('objects')
        df = pd.DataFrame.from_dict(nodes)

        def convert_units(v, to_unit='Mi'):
            # convert a value like '452456Ki' to '452Mi'
            # https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/
            CONVERSION = {
                'Ki': 1024,
                'Mi': 1024 ** 2,
            }
            if v[-2:] in CONVERSION:
                value, unit = int(v[0:-2]), v[-2:]
                value = int(value * CONVERSION[unit] * 1 / CONVERSION[to_unit])
            else:
                # assume bytes
                value = int(v) * 1
                to_unit = ''
            return f'{value}{to_unit}'

        def node_summary(row):
            capacity = {k: convert_units(v) for k, v in row['capacity'].items()}
            status = row['status']
            row['cpu'] = capacity['cpu']
            row['memory'] = f"{capacity['memory']}"
            row['disk'] = f"{capacity['ephemeral-storage']}"
            row['status'] = 'running' if status['Ready'] == 'True' else 'not ready'
            return row

        df = df.apply(node_summary, axis=1)
        cols = 'name,status,role,cpu,memory,disk'.split(',')
        print(tabulate(df[cols], headers='keys', showindex=False))

    def status_dbsize(self, kind, auth):
        data = self._get_status(kind, auth)
        dbsize = data.get('objects')
        df = pd.DataFrame([dbsize])
        cols = ['kind', 'size', 'status']
        print(tabulate(df[cols], headers='keys', showindex=False))

    def log(self):
        om = self.om
        podname = self.args.get('<pod>', 'missing')
        since = self.args.get('--since', '5m')
        auth = OmegaRestApiAuth.make_from(om)
        data = self._get_logs(podname, since, auth)
        entries = data.get('entries')
        print(entries)


def prom2df(data, metric_name):
    """ convert prom response to pandas dataframe
    """
    import pandas as pd
    def parse_vector(metrics):
        for item in metrics:
            data = item['metric']
            values = [item.get('value')] if 'value' in item else item.get('values')
            for ts, value in values:
                entry = dict(**data, ts=ts, value=float(value))
                yield entry

    metrics = data['data']['result']
    df = pd.DataFrame(parse_vector(metrics))
    df['ts'] = pd.to_datetime(df['ts'], unit='s')
    df['metric_name'] = metric_name
    base_cols = ['metric_name', 'ts']
    cols = base_cols + [c for c in df.columns if c not in base_cols]
    return df[cols]



import numpy as np
import pandas as pd

from omegaml.util import ensure_json_serializable


class GenericModelResource(object):
    def __init__(self, om, is_async=False):
        self.om = om
        self.is_async = is_async

    def is_eager(self):
        return getattr(self.om.runtime.celeryapp.conf, 'CELERY_ALWAYS_EAGER', False)

    def metadata(self, request, name):
        meta = self.om.models.metadata(name)
        data = {
            'model': {
                'name': meta.name,
                'kind': meta.kind,
                'created': '{}'.format(meta.created),
                'bucket': meta.bucket,
            }
        }
        return data

    def predict(self, model_id, query, payload):
        """
        Args:
            model_id (str): the name of the model
            query (dict): the query parameters
            payload (dict): the json body

        Returns:
            dict(model: id, result: dict)
        """
        data = (payload or {}).get('data')
        dataset = query.get('datax')
        if data is not None:
            columns = payload.get('columns')
            shape = payload.get('shape')
            df = pd.DataFrame(data)[columns]
            if shape is not None:
                assert len(columns) == 1, "only 1 column allowed to be reshaped"
                col = columns[0]
                df[col] = df[col].apply(lambda v: np.array(v).reshape(shape))
                df = np.stack(df[col])
            promise = self.om.runtime.model(model_id).predict(df)
        elif dataset:
            promise = self.om.runtime.model(model_id).predict(dataset)
        else:
            raise ValueError('require either "data" key in body, or ?datax=dataset')
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def prepare_result(self, result, model_id=None, **kwargs):
        return {'model': model_id, 'result': ensure_json_serializable(result)}

    def fit(self, model_id, query, payload):
        datax = query.get('datax')
        datay = query.get('datay')
        promise = self.om.runtime.model(model_id).fit(datax, datay)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def predict_proba(self, model_id, query, payload):
        datax = query.get('datax')
        datay = query.get('datay')
        promise = self.om.runtime.model(model_id).predict_proba(datax, datay)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def partial_fit(self, model_id, query, payload):
        datax = query.get('datax')
        datay = query.get('datay')
        promise = self.om.runtime.model(model_id).partial_fit(datax, datay)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def score(self, model_id, query, payload):
        datax = query.get('datax')
        datay = query.get('datay')
        promise = self.om.runtime.model(model_id).score(datax, datay)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def transform(self, model_id, query, payload):
        datax = query.get('datax')
        promise = self.om.runtime.model(model_id).transform(datax)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result

    def decision_function(self, model_id, query, payload):
        datax = query.get('datax')
        promise = self.om.runtime.model(model_id).decision_function(datax)
        result = self.prepare_result(promise.get(), model_id=model_id) if not self.is_async else promise
        return result





from __future__ import absolute_import

import random
import string

import numpy as np
import os
import pandas as pd
import unittest
from pandas.util.testing import assert_frame_equal, assert_series_equal
from six.moves import range
from unittest.case import TestCase, skip

from omegaml import Omega
from omegaml.mdataframe import MDataFrame
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import flatten_columns


class MDataFrameTests(OmegaTestMixin, TestCase):

    def setUp(self):
        TestCase.setUp(self)
        df = self.df = pd.DataFrame({'x': list(range(0, 10)) + list(range(0, 10)),
                                     'y': random.sample(list(range(0, 100)), 20)})
        om = self.om = Omega()
        self.clean()
        om.datasets.put(df, 'sample', append=False)
        self.coll = om.datasets.collection('sample')

    def tearDown(self):
        TestCase.tearDown(self)

    def test_groupby(self):
        coll = self.coll
        df = self.df
        keys = []
        for key, groupdf in MDataFrame(coll).groupby(['x']):
            x = key.get('x')
            keys.append(x)
            subdf = df[df.x == x]
            assert_frame_equal(subdf, groupdf.value)
        self.assertEqual(set(keys), set(df.x))

    def test_count(self):
        coll = self.coll
        df = self.df
        counts = MDataFrame(coll).groupby(['x']).count()
        test_counts = df.groupby('x').count()
        self.assertTrue(test_counts.equals(counts))

    def test_count_multi_columns(self):
        coll = self.coll
        df = self.df
        # add a column
        mdf = MDataFrame(coll)
        mdf['z'] = 5
        df['z'] = 5
        # group by and count
        counts = mdf.groupby(['x']).count()
        test_counts = df.groupby('x').count()
        self.assertTrue(test_counts.equals(counts))

    def test_count_column(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).groupby(['x']).x.count()
        testgroup = df.groupby('x').x.count()
        self.assertTrue(result.equals(testgroup))

    def test_aggregate(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).groupby(['x']).agg({'x': 'sum'})
        testagg = df.groupby('x').agg({'x': 'sum'})
        testagg = testagg.rename(columns=dict(x='x_sum'))
        self.assertTrue(result.equals(testagg))

    def test_aggregate_multi_stats(self):
        coll = self.coll
        df = self.df
        stats = {'x': ['sum', 'mean', 'max', 'min', 'std']}
        result = MDataFrame(coll).groupby(['x']).agg(stats)
        testagg = df.groupby('x').agg(stats)
        testagg.columns = testagg.columns.map(flatten_columns)
        testagg = testagg[result.columns]
        assert_frame_equal(testagg, result, check_dtype=False)

    def test_mdataframe(self):
        coll = self.coll
        df = self.df
        mdf = MDataFrame(coll)
        result = mdf.value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))
        self.assertEqual(mdf.shape, df.shape)

    def test_mdataframe_count(self):
        coll = self.coll
        df = self.df
        mdf = MDataFrame(coll)
        assert_series_equal(df.count(), mdf.count())
        self.assertEqual(len(mdf), len(mdf))

    def test_mdataframe_xlarge(self):
        df = pd.DataFrame({
            'a': list(range(0, int(1e6 + 1))),
            'b': list(range(0, int(1e6 + 1)))
        })
        store = self.om.datasets
        store.put(df, 'mydata-xlarge', append=False)
        coll = store.collection('mydata-xlarge')
        result = MDataFrame(coll).value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))

    def test_mdataframe_column_attribute(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).y.value
        self.assertTrue(df.y.equals(result))

    def test_mdataframe_column_slice(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll)['y'].value
        self.assertTrue(df['y'].equals(result))

    def test_mdataframe_columns_slice(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll)[['x', 'y']].value
        self.assertTrue(df[['x', 'y']].equals(result))

    def test_mdataframe_sort(self):
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).sort(['-x', '-y']).value
        df = df.sort_values(
            ['x', 'y'], ascending=[False, False])
        assert_frame_equal(df, result)

    def test_mdataframe_merge(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 20)),
                              'y': list(range(0, 20)),
                              'z': list(range(0, 20))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_differing_columns(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'s': list(range(0, 20)),
                              'y': list(range(0, 20)),
                              'z': list(range(0, 20))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, left_on='x',
                                        right_on='s', how='left').value
        testdf = df.merge(other, left_on='x', right_on='s', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_nomatch(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(50, 55)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    @unittest.skip("disalbed because MDataFrame.append fails and is very slow")
    def test_mdataframe_merge_append(self):
        ## FIXME this does not work
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        mdf = om.datasets.getl('samplez')
        mdf.append(mdf)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left', suffixes=('', '')).value
        testdf = df.append(other, ignore_index=True)
        testdf = testdf[result.columns]
        assert_frame_equal(result, testdf)

    def test_mdataframe_merge_partial_match(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left').value
        testdf = df.merge(other, on='x', how='left')
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_inner(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='inner',
                                        sort=True).value
        testdf = df.merge(other, on='x', how='inner', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_right(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left',
                                        sort=True).value
        testdf = df.merge(other, on='x', how='left', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_right_cartesian(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        om.datasets.put(other, 'samplez', append=True)
        other = om.datasets.get('samplez')
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left',
                                        sort=True).value
        testdf = df.merge(other, on='x', how='left', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_mdataframe_merge_filtered(self):
        coll = self.coll
        df = self.df
        om = self.om
        other = pd.DataFrame({'x': list(range(0, 5)),
                              'y': list(range(0, 5)),
                              'z': list(range(0, 5))})
        om.datasets.put(other, 'samplez', append=False)
        om.datasets.put(other, 'samplez', append=True)
        other = om.datasets.get('samplez')
        coll2 = om.datasets.collection('samplez')
        result = MDataFrame(coll).merge(coll2, on='x', how='left',
                                        sort=True, filter=dict(x__in=[1, 2])).value
        q = df['x'].isin([1, 2])
        testdf = df[q].merge(other, on='x', how='left', sort=True)
        testdf = testdf[result.columns]
        self.assertTrue(result.equals(testdf))

    def test_verylarge_dataframe(self):
        if not os.environ.get('TEST_LARGE'):
            return
        other = pd.DataFrame({'x': list(range(0, int(10e6))),
                              'y': list(range(0, int(10e6))),
                              'z': list(range(0, int(10e6)))})
        coll = self.coll
        df = self.df
        result = MDataFrame(coll).value
        self.assertEqual(set(MDataFrame(coll).columns),
                         set(list(df.columns)))
        self.assertTrue(result.equals(df))

    def test_unique_series(self):
        coll = self.coll
        df = self.df
        om = self.om
        om.datasets.put(df, 'uniques', append=False)
        coll = om.datasets.collection('uniques')
        result = MDataFrame(coll).x.unique().value
        self.assertListEqual(list(result), list(df.x.unique()))

    def test_query_null(self):
        om = self.om
        df = pd.DataFrame({'x': list(range(0, 5)),
                           'y': [1, 2, 3, None, None]})
        om.datasets.put(df, 'foox', append=False)
        result = om.datasets.get('foox', y__isnull=True, lazy=True).value
        test = df[df.isnull().any(axis=1)]
        assert_frame_equal(result, test)

    def test_query_pandas_style(self):
        om = self.om
        df = pd.DataFrame({'x': list(range(0, 5)),
                           'y': [1, 2, 3, None, None]})
        om.datasets.put(df, 'foox', append=False)
        mdf = om.datasets.getl('foox')
        # simple subset
        mdf_flt = mdf['x'] == 4
        df_flt = df['x'] == 4
        assert_frame_equal(mdf[mdf_flt].value, df[df_flt])
        # and combined
        mdf_flt = (mdf['x'] == 4) & (mdf['x'] < 5)
        df_flt = (df['x'] == 4) & (df['x'] < 5)
        assert_frame_equal(mdf[mdf_flt].value, df[df_flt])
        # or combined
        mdf_flt = (mdf['x'] < 3) | (mdf['x'] > 3)
        df_flt = (df['x'] < 3) | (df['x'] > 3)
        assert_frame_equal(mdf[mdf_flt].value, df[df_flt])
        # negative combined
        mdf_flt = (mdf['x'] < 3) | (mdf['x'] > 3)
        df_flt = (df['x'] < 3) | (df['x'] > 3)
        assert_frame_equal(mdf[~mdf_flt].value, df[~df_flt])
        # partial negative combined

    @skip
    def test_partial_negative_query(self):
        om = self.om
        df = pd.DataFrame({'x': list(range(0, 5)),
                           'y': [1, 2, 3, None, None]})
        om.datasets.put(df, 'foox', append=False)
        mdf = om.datasets.getl('foox')
        # TODO this fails and should not
        mdf_flt = (~(mdf['x'] < 3)) | (mdf['x'] > 3)
        df_flt = (~(df['x'] < 3)) | (df['x'] > 3)
        assert_frame_equal(mdf[mdf_flt].value, df[df_flt])

    def test_locindexer_numeric_index(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc[4].value
        assert_series_equal(df.loc[4], dfx)
        # by slice
        dfx = om.datasets.getl('foo').loc[2:4].value
        assert_frame_equal(df.loc[2:4], dfx)
        # by list
        dfx = om.datasets.getl('foo').loc[[2, 4]].value
        assert_frame_equal(df.loc[[2, 4]], dfx)
        # by ndarray
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo').loc[sel, :].value
        assert_frame_equal(df.loc[sel, :], dfx, check_names=False)

    def test_locindexer_character_index(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc['c'].value
        assert_series_equal(df.loc['c'], dfx)
        # by slice
        dfx = om.datasets.getl('foo').loc['c':'f'].value
        assert_frame_equal(df.loc['c':'f'], dfx)
        # by list
        dfx = om.datasets.getl('foo').loc[['c', 'f']].value
        assert_frame_equal(df.loc[['c', 'f']], dfx)

    def test_locindexer_timeseries_index(self):
        om = self.om
        # create some dataframe
        tsidx = pd.date_range(pd.datetime(2016, 1, 1), pd.datetime(2016, 4, 1))
        df = pd.DataFrame({
            'a': list(range(0, len(tsidx))),
            'b': list(range(0, len(tsidx)))
        }, index=tsidx)
        om.datasets.put(df, 'foo', append=False)
        # by label
        dfx = om.datasets.getl('foo').loc[pd.datetime(2016, 2, 3)].value
        assert_series_equal(dfx, df.loc[pd.datetime(2016, 2, 3)])
        # by slice
        start, end = pd.datetime(2016, 2, 3), pd.datetime(2016, 2, 8)
        dfx = om.datasets.getl('foo').loc[start:end].value
        assert_frame_equal(df.loc[start:end], dfx)

    def test_locindexer_multiindex(self):
        # create some dataframe
        om = self.om
        midx = pd.MultiIndex(levels=[[u'bar', u'baz', u'foo', u'qux'],
                                     [u'one', u'two']],
                             codes=[
                                 [0, 0, 1, 1, 2, 2, 3, 3],
                                 [0, 1, 0, 1, 0, 1, 0, 1]],
                             names=[u'first', u'second'])
        df = pd.DataFrame({'x': range(0, len(midx))}, index=midx)
        om.datasets.put(df, 'foomidx', append=False)
        dfx = om.datasets.getl('foomidx').loc['bar', 'one'].value
        assert_series_equal(dfx, df.loc['bar', 'one'])

    def test_locindexer_series(self):
        """ test storing a pandas series with it's own index """
        om = self.om
        series = pd.Series(range(10),
                           name='foo',
                           index=pd.date_range(pd.datetime(2016, 1, 1),
                                               pd.datetime(2016, 1, 10)))
        om.datasets.put(series, 'fooseries', append=False)
        # try data range
        daterange = slice(pd.datetime(2016, 1, 5), pd.datetime(2016, 1, 10))
        series2 = om.datasets.getl('fooseries').loc[daterange].value
        assert_series_equal(series2, series.loc[daterange])
        # try single date
        daterange = pd.datetime(2016, 1, 5)
        series2 = om.datasets.getl('fooseries').loc[daterange].value
        self.assertEqual(series2, series.loc[daterange])

    def test_ilocindexer(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0].value
        assert_series_equal(df.iloc[0], dfx)
        # by slice
        dfx = om.datasets.getl('foo').iloc[0:1].value
        assert_frame_equal(df.iloc[0:1], dfx)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2]].value
        assert_frame_equal(df.iloc[[1, 2]], dfx)

    def test_ilocindexer_single_column(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0, 1].value
        self.assertEqual(df.iloc[0, 1], dfx)
        # by slice
        # FIXME column access by iloc is not guaranteed to return in order
        dfx = om.datasets.getl('foo').iloc[0:2, 1].value
        assert_series_equal(df.iloc[0:2, 1], dfx, check_names=False)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2], 1].value
        assert_series_equal(df.iloc[[1, 2], 1], dfx, check_names=False)

    def test_ilocindexer_columns(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by single location
        dfx = om.datasets.getl('foo').iloc[0, :].value
        assert_series_equal(df.iloc[0, :], dfx)
        # by slice
        # FIXME column access by iloc is not guaranteed to return in order
        dfx = om.datasets.getl('foo').iloc[0:2, :].value
        assert_frame_equal(df.iloc[0:2, :], dfx, check_names=False)
        # by list
        dfx = om.datasets.getl('foo').iloc[[1, 2], :].value
        assert_frame_equal(df.iloc[[1, 2], :], dfx, check_names=False)
        # by ndarray
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo').iloc[sel, :].value
        assert_frame_equal(df.iloc[sel, :], dfx, check_names=False)

    def test_ilocindexer_array(self):
        om = self.om
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        idx = string.ascii_lowercase[0:9]
        df = pd.DataFrame(data, index=(c for c in idx))
        om.datasets.put(df, 'foo', append=False)
        # by ndarray with immediate loc
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo')
        dfx.immediate_loc = True
        dfx = dfx[['a']].iloc[sel]
        assert_frame_equal(df[['a']].iloc[sel], dfx, check_names=False)
        # by ndarray with delayed loc
        sel = np.array([1, 2])
        dfx = om.datasets.getl('foo')
        dfx.immediate_loc = False
        dfx = dfx[['a']].iloc[sel].value
        assert_frame_equal(df[['a']].iloc[sel], dfx, check_names=False)

    def test_iterchunks(self):
        om = self.om
        for cdf in om.datasets.getl('sample', chunksize=2):
            self.assertEqual(len(cdf), 2)

    def test_iterrows(self):
        om = self.om
        mdf = om.datasets.getl('sample')
        df = mdf.value
        for df_row, mdf_row in zip(mdf.iterrows(), df.iterrows()):
            self.assertEqual(type(df_row), type(mdf_row))
            assert_series_equal(df_row[1], mdf_row[1])

    def test_iteritems(self):
        om = self.om
        mdf = om.datasets.getl('sample')
        df = mdf.value
        for df_row, mdf_row in zip(mdf.iteritems(), df.iteritems()):
            self.assertEqual(type(df_row), type(mdf_row))
            assert_series_equal(df_row[1], mdf_row[1])

    def test_items(self):
        om = self.om
        mdf = om.datasets.getl('sample')
        df = mdf.value
        for df_row, mdf_row in zip(mdf.items(), df.items()):
            self.assertEqual(type(df_row), type(mdf_row))
            assert_series_equal(df_row[1], mdf_row[1])



import datetime
import re

import flask
import numpy as np
import pandas as pd
import six
from flask import request
from flask_restplus import Resource, fields
from mongoengine import DoesNotExist
from werkzeug.exceptions import NotFound, BadRequest

import omegaml as om
from omegaml.backends.restapi.asyncrest import AsyncTaskResourceMixin, AsyncResponseMixin, resolve
from omegaml.backends.restapi.job import GenericJobResource
from omegaml.backends.restapi.model import GenericModelResource
from omegaml.backends.restapi.script import GenericScriptResource
from omegaml.restapi.util import strict
from .app import api

isTrue = lambda v: v if isinstance(v, bool) else (
      v.lower() in ['yes', 'y', 't', 'true', '1'])

PredictInput = strict(api).model('ModelInputSchema', {
    'columns': fields.List(fields.String),
    'data': fields.List(fields.Raw),
    'shape': fields.List(fields.Integer),
})

PredictOutput = api.model('PredictOutput', {
    'model': fields.String,
    'result': fields.Raw,
    'resource_uri': fields.String,
})

TaskInput = strict(api).model('TaskInput', {
    'resource_uri': fields.String,
})

TaskOutput = strict(api).model('TaskOutput', {
    'task_id': fields.String,
    'status': fields.String,
    'response': fields.Raw
})

DatasetInput = api.model('DatasetInput', {
    'data': fields.Raw,
    'dtypes': fields.Raw,
    'append': fields.Boolean,
})

DatasetIndex = api.model('DatasetIndex', {
    'values': fields.List(fields.Raw),
    'type': fields.String,
})

DatasetQueryOutput = api.model('DatasetQueryOutput', {
    'data': fields.Raw,
    'index': fields.Nested(DatasetIndex)
})

ScriptInput = api.model('ScriptInput', {
})

ScriptOutput = api.model('ScriptOutput', {
    'resource_uri': fields.String,
    'script': fields.String,
    'result': fields.Raw,
    'runtimes': fields.Float,
    'started': fields.DateTime,
    'ended': fields.DateTime,
})

JobInput = api.model('JobInput', {
})

JobOutput = api.model('JobOutput', {
    'resource_uri': fields.String,
    'job': fields.String,
    'source_job': fields.String,
    'job_results': fields.Raw,
    'job_runs': fields.List(fields.Raw),
    'created': fields.DateTime,
})


class OmegaResourceMixin(object):
    """
    helper mixin to resolve the request to a configured Omega instance
    """

    def __init__(self, *args, **kwargs):
        self._omega_instance = None
        super().__init__(*args, **kwargs)

    def dispatch_request(self, *args, **kwargs):
        self._omega_instance = None  # always start with a fresh omega instance
        return super().dispatch_request(*args, **kwargs)

    @property
    def _omega(self):
        if self._omega_instance is None:
            bucket = flask.request.headers.get('bucket')
            self._omega_instance = om.setup()[bucket]
        return self._omega_instance

    def get_query_payload(self):
        query = flask.request.args.to_dict()
        payload = api.payload or {}
        return query, payload

    def create_response_from_resource(self, generic_resource, resource_method, resource_name, resource_pk, *args,
                                      **kwargs):
        query, payload = self.get_query_payload()
        async_body = {
            resource_name: resource_pk,
            'result': 'pending',
        }
        try:
            meth = self._get_resource_method(generic_resource, resource_method)
            result = meth(resource_pk, query, payload)
            resp = self.create_maybe_async_response(result, async_body=async_body, **kwargs)
        except Exception as e:
            raise BadRequest(str(e))
        return resp

    def _get_resource_method(self, resource_name, method_name):
        resource = getattr(self, resource_name)
        meth = getattr(resource, method_name)
        return meth

    @property
    def _generic_model_resource(self):
        return GenericModelResource(self._omega, is_async=self.is_async)

    @property
    def _generic_script_resource(self):
        return GenericScriptResource(self._omega, is_async=self.is_async)

    @property
    def _generic_job_resource(self):
        return GenericJobResource(self._omega, is_async=self.is_async)

    @property
    def celeryapp(self):
        return self._omega.runtime.celeryapp


@api.route('/api/v1/ping')
class PingResource(Resource):
    def get(self):
        dt = datetime.datetime.now()
        return {'ping': dt.isoformat()}


@api.route('/api/v1/task/<string:taskid>/<string:action>')
class TaskResource(OmegaResourceMixin, AsyncTaskResourceMixin, Resource):
    @api.expect(TaskInput, validate=False)
    @api.marshal_with(TaskOutput)
    def get(self, taskid, action):
        return self.process_task_action(taskid, action, request=flask.request)

    def resolve_resource_method(self, taskid, context):
        # this maps a tasks resource_uri to a resource method to resolve the async result
        resourceUri = api.payload.get('resource_uri')
        path, res_kwargs = resolve(resourceUri)
        URI_METHOD_MAP = {
            r"/api/.*/model/.*/.*/?$": ('_generic_model_resource', 'prepare_result', {'pk': 'model_id'}),
            r'/api/.*/job/.*/run/?$': ('_generic_job_resource', 'prepare_result_from_run', {'pk': 'job_id'}),
            r'/api/.*/script/.*/run/?$': ('_generic_script_resource', 'prepare_result_from_run', {'pk': 'script_id'}),
        }
        for regexp, specs in URI_METHOD_MAP.items():
            if re.match(regexp.replace(r'/', r'\/'), resourceUri):
                resource_name, method_name, kwargs_map = specs
                meth = self._get_resource_method(resource_name, method_name)
                return meth, res_kwargs
        raise NotFound('unknown resource {}'.format(resourceUri))


@api.route('/api/v1/model/<path:model_id>/<string:action>', methods=['GET', 'PUT'])
@api.route('/api/v1/model/<path:model_id>/', defaults={'action': 'metadata'}, methods=['GET'])
class ModelResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(PredictInput, validate=False)
    @api.marshal_with(PredictOutput)
    def get(self, model_id, action=None):
        return self.create_response_from_resource('_generic_model_resource', action, 'model', model_id)

    @api.expect(PredictInput, validate=False)
    @api.marshal_with(PredictOutput)
    def put(self, model_id, action=None):
        return self.create_response_from_resource('_generic_model_resource', action, 'model', model_id)


@api.route('/api/v1/script/<path:script_id>/run', methods=['POST'])
class ScriptResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(ScriptInput, validate=False)
    @api.marshal_with(ScriptOutput)
    def post(self, script_id):
        return self.create_response_from_resource('_generic_script_resource', 'run', 'script', script_id)


@api.route('/api/v1/job/<path:job_id>/run', methods=['POST'])
@api.route('/api/v1/job/<path:job_id>/', methods=['GET'])
@api.route('/api/v1/job/', methods=['GET'], defaults={'job_id': None})
class JobResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(JobInput, validate=False)
    @api.marshal_with(JobOutput)
    def get(self, job_id):
        action = 'list' if job_id is None else 'metadata'
        return self.create_response_from_resource('_generic_job_resource', action, 'job', job_id)

    @api.expect(JobInput, validate=False)
    @api.marshal_with(JobOutput)
    def post(self, job_id):
        return self.create_response_from_resource('_generic_job_resource', 'run', 'job', job_id)


@api.route('/api/v1/dataset/<path:dataset_id>')
class DatasetResource(OmegaResourceMixin, Resource):
    def _restore_filter(self, om, fltparams, name):
        """
        restore filter kwargs for query in om.datasets.get
        """
        # -- get filters as specified on request query args
        fltkwargs = {k: v for k, v in six.iteritems(fltparams)
                     if k not in ['orient', 'limit', 'skip', 'page']}
        # -- get dtypes of dataframe and convert filter values
        metadata = om.datasets.metadata(name)
        kind_meta = metadata.kind_meta or {}
        dtypes = kind_meta.get('dtypes')
        # get numpy/python typemap. this is required for Py3 support
        # adopted from https://stackoverflow.com/a/34919415
        np_typemap = {v: getattr(six.moves.builtins, k)
                      for k, v in np.typeDict.items()
                      if k in vars(six.moves.builtins)}
        for k, v in six.iteritems(fltkwargs):
            # -- get column name without operator (e.g. x__gt => x)
            col = k.split('__')[0]
            value = six.moves.urllib.parse.unquote(v)
            dtype = dtypes.get(str(col))
            if dtype:
                # -- get dtyped value and convert to python type
                value = np_typemap.get(getattr(np, dtype, str), str)(value)
                fltkwargs[k] = value
        return fltkwargs

    @api.marshal_with(DatasetQueryOutput)
    def get(self, dataset_id):
        om = self._omega
        orient = request.args.get('orient', 'dict')
        fltkwargs = self._restore_filter(om, request.args, dataset_id)
        df = om.datasets.getl(dataset_id, filter=fltkwargs).value
        # get index values as python types to support Py3
        index_values = list(df.index.astype('O').values)
        index_type = type(df.index).__name__
        # get data set values as python types to support Py3
        df = df.reset_index(drop=True)
        df.index = df.index.astype('O')
        # convert nan to None
        # https://stackoverflow.com/a/34467382
        data = df.where(pd.notnull(df), None).astype('O').to_dict(orient)
        return {
            'data': data,
            'index': {
                'values': index_values,
                'type': index_type,
            }
        }

    @api.expect(DatasetInput, validate=True)
    @api.response(200, 'updated')
    def put(self, dataset_id):
        om = self._omega
        orient = request.args.get('orient', 'dict')
        if 'append' in request.args:
            append = isTrue(request.args.get('append', 'true'))
        else:
            append = isTrue(api.payload.get('append', 'true'))
        dtypes = api.payload.get('dtypes')
        if orient == 'dict':
            orient = 'columns'
        if dtypes:
            # due to https://github.com/pandas-dev/pandas/issues/14655#issuecomment-260736368
            dtypes = {k: np.dtype(v) for k, v in six.iteritems(dtypes)}
        df = pd.DataFrame.from_dict(api.payload.get('data'),
                                    orient=orient).astype(dtypes)
        om.datasets.put(df, dataset_id, append=append)
        return '', 200

    @api.response(200, 'updated')
    @api.response(404, 'does not exist')
    def delete(self, dataset_id):
        om = self._omega
        try:
            om.datasets.drop(dataset_id)
        except DoesNotExist:
            status = 404
        else:
            status = 200
        return '', status



from os.path import basename, dirname

import os
import six

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.package.packager import build_sdist, install_and_import, load_from_path


class PythonPackageData(BaseDataBackend):
    """
    Backend to support locally sourced custom scripts deployment to runtimes cluster

    This supports any local setup.py

    Usage:
        om.scripts.put('pkg://path/to/setup.py')
    """
    KIND = 'python.package'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, six.string_types) and obj.startswith('pkg://')

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a python package

        :param obj: full path to package file or directory, sytanx as
                    pkg://path/to/dist.tar.gz or pkg://path/to/setup.py
        :param name: name of package. must be the actual name of the package
                     as specified in setup.py
        :return: the Metadata object
        """
        pkgsrc = pkgdist = obj.split('//')[1]
        if not 'tar.gz' in os.path.basename(pkgdist):
            pkgsrc = pkgsrc.replace('setup.py', '')
            distdir = os.path.join(pkgsrc, 'dist')
            sdist = build_sdist(pkgsrc, distdir)
            version = sdist.metadata.version
            pkgname = sdist.metadata.name
            pkgdist = os.path.join(distdir, '{pkgname}-{version}.tar.gz'.format(**locals()))
        filename = self.data_store.object_store_key(name, 'pkg', hashed=True)
        gridfile = self._store_to_file(self.data_store, pkgdist, filename)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=PythonPackageData.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, keep=False, **kwargs):
        """
        Load package from store, install it locally and load.

        :param name: the name of the package
        :param keep: keep the packages load path in sys.path, defaults to False
        :param kwargs:
        :return: the loaded module
        """
        pkgname = basename(name)
        packagefname = '{}.tar.gz'.format(os.path.join(self.data_store.tmppath, pkgname))
        os.makedirs(dirname(packagefname), exist_ok=True)
        self.path = self.packages_path
        dstdir = self.path
        if not os.path.exists(os.path.join(dstdir, pkgname)):
            meta = self.data_store.metadata(name)
            outf = meta.gridfile
            with open(packagefname, 'wb') as pkgf:
                pkgf.write(outf.read())
            mod = install_and_import(packagefname, pkgname, dstdir, keep=keep)
        else:
            mod = load_from_path(pkgname, dstdir, keep=keep)
        return mod

    @property
    def packages_path(self):
        return os.path.join(self.data_store.tmppath, 'packages')



from traitlets.config import Config

from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class ShellCommandBase(CommandBase):
    """
    Usage:
        om shell [<command>] [options]

    Description:
        Without a command will start an IPython shell with omega|ml ready to use

        $ om shell
        [] om.runtime.ping()
        => { ... }

        By passing a command, run arbitrary Python code

        $ om shell "om.runtime.ping()"
    """
    command = 'shell'

    def shell(self):
        use_ipython = False
        command = self.args.get('<command>') or ''
        try:
            import IPython
        except:
            self.logger.warn("you should pip install ipython for convenience")
        else:
            use_ipython = True
        # ipython
        if use_ipython:
            c = Config()
            c.InteractiveShellApp.exec_lines = [
                'from omegaml.client.util import get_omega',
                'om = get_omega(shell_args)',
                'print("omegaml is available as the om variable")',
            ]
            c.TerminalIPythonApp.display_banner = False
            argv = f'-c {command}'.split(' ') if command else []
            IPython.start_ipython(argv,
                                  config=c, user_ns=dict(shell_args=self.args))
            return
        # default console
        import code
        om = get_omega(self.args)
        try:
            import gnureadline
        except:
            self.logger.warn("you should pip install gnureadline for convenience")
        variables = {}
        variables.update(locals())
        shell = code.InteractiveConsole(locals=variables)
        shell.interact()
